import pathlib
import bs4
import numpy as np
import re
import gzip
import requests

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog


class Downloader(object):

    def __init__(
            self,
            direct_download=True,
            force_download=False,
    ):

        self._url_base = 'https://www.ncei.noaa.gov/data/dmsp-space-weather-sensors/access/'
        self.force_download = force_download
        self.done = False
        if direct_download:
            self.download()

    def download(self, url, file_dir=None, file_name=None):

        file_path_remote = pathlib.Path(url)
        file_name_remote = file_path_remote.name

        if file_name is None:
            if file_path_remote.suffix == '.gz':
                file_name = file_path_remote.with_suffix('').name
            else:
                file_name = file_name_remote

        file_path_local = file_dir / file_name
        if file_path_local.is_file() and not self.force_download:
            mylog.simpleinfo.info(
                "The file {} exists in {}!".format(file_path_local.name, file_dir)
            )
            return file_path_local

        mylog.simpleinfo.info("Contacting NCEI ...")
        content = self._get_content(url)

        if file_path_remote.suffix == '.gz':
            content = gzip.decompress(content)

        with open(file_path_local, 'wb') as f:
            f.write(content)
        mylog.simpleinfo.info("The file {} has been downloaded to {}.".format(
            file_path_local.name, file_dir
        ))
        return file_path_local

    @staticmethod
    def search_files(url, file_name_patterns=None):
        if file_name_patterns is None:
            file_name_patterns = []
        r = requests.get(url)
        r.raise_for_status()
        soup = bs4.BeautifulSoup(r.text, 'html.parser')
        links = soup.find_all('a', href=True)
        r.close()

        files = []
        search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
        rc = re.compile(search_pattern)
        for link in links:
            href = link['href']
            if not any(href.endswith(s) for s in ['.gz', '.cdf', '.dat', '.txt', '.pdf', '.png']):
                continue

            file_name = href.split('/')[-1]
            res = rc.match(file_name)
            if res is None:
                continue
            file_url = url + '/' + href
            files.append(file_url)

        if not list(files):
            mylog.StreamLogger.warning("No files matching the reqeust from {}!".format(url))
            return None

        return files


    @staticmethod
    def _get_content(url):
        r = requests.get(url)
        r.raise_for_status()
        content = r.content
        r.close()
        return content




import numpy as np
import datetime




def get_sub_solar_point(dts):


    ye, mo, da, ho, mi, se = utc
    ta = pi * 2
    ut = ho + mi / 60 + se / 3600
    t = 367 * ye - 7 * (ye + (mo + 9) // 12) // 4
    dn = t + 275 * mo // 9 + da - 730531.5 + ut / 24
    sl = dn * 0.01720279239 + 4.894967873
    sa = dn * 0.01720197034 + 6.240040768
    t = sl + 0.03342305518 * sin(sa)
    ec = t + 0.0003490658504 * sin(2 * sa)
    ob = 0.4090877234 - 0.000000006981317008 * dn
    st = 4.894961213 + 6.300388099 * dn
    ra = atan2(cos(ob) * sin(ec), cos(ec))
    de = asin(sin(ob) * sin(ec))
    la = degrees(de)
    lo = degrees(ra - st) % 360
    lo = lo - 360 if lo > 180 else lo
    return [round(la, 6), round(lo, 6)]




def solar_declination(dt, degrees=False):
    factor = np.pi / 180.
    sd_SS = - 23.4333333 * factor   # Declination angle at southern summer solstice.

    dt_0 = datetime.datetime(dt.year, dt.month, dt.day, 0, )
    N = (dt - dt_0).total_seconda() / 86400.
    sin_EL = np.cos(
        2 * np.pi / 365.24 * (N + 10) + 2 * 0.0167 * np.sin(2 * np.pi/365.24 * (N - 2))
    )

    sd = np.arcsin(
        np.sin(sd_SS) * sin_EL
    )

    if degrees:
        sd = sd / factor
    return sd


def subsolar_point(dt, degrees=False):
    phi = solar_declination(dt, degrees)

    lamda = - 15 * ((dt.hour + dt.minute / 60 + dt.second / 3600) - 12. + equation_of_time(dt))

    
def equation_of_time(dt):
    return




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, ProductModel
from geospacelab.datahub.sources.wdc import wdc_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.sources.wdc.dst.loader import Loader as default_Loader
from geospacelab.datahub.sources.wdc.dst.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.wdc.dst.variable_config as var_config


default_dataset_attrs = {
    'database': wdc_database,
    'product': 'Dst',
    'data_file_ext': 'nc',
    'data_root_dir': prf.datahub_data_root_dir / 'WDC' / 'Dst',
    'allow_load': True,
    'allow_download': True,
    'data_search_recursive': False,
    'label_fields': ['database', 'product', 'data_file_ext'],
    'time_clip': True,
}

default_variable_names = ['DATETIME', 'Dst']

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', '')
        self.product = kwargs.pop('product', '')
        self.allow_download = kwargs.pop('allow_download', True)

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type=self.data_file_ext)

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def search_data_files(self, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_months = dttool.get_diff_months(dt_fr, dt_to)
        dt0 = dttool.get_first_day_of_month(dt_fr)
        for i in range(diff_months + 1):
            thismonth = dttool.get_next_n_months(dt0, i)

            initial_file_dir = kwargs.pop('initial_file_dir', None)
            if initial_file_dir is None:
                initial_file_dir = self.data_root_dir / thismonth.strftime("%Y")
            file_patterns = [thismonth.strftime("%Y%m")]
            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir, search_pattern=search_pattern
            )

            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern)
                else:
                    print('Cannot find files from the online database!')

        return done

    def download_data(self):
        if self.data_file_ext == 'nc':
            download_obj = self.downloader(self.dt_fr, self.dt_to, data_file_root_dir=self.data_root_dir)
        else:
            raise NotImplementedError
        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError














# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

from cProfile import label
import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as dtparse
import numpy as np
import re
import requests
import bs4
import os
import pathlib

from geospacelab.config import prf
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool
from geospacelab.datahub.sources.madrigal.downloader import Downloader as DownloaderBase


def test():
    dt_fr = datetime.datetime(2016, 3, 14)
    dt_to = datetime.datetime(2016, 3, 14, 23)
    eid = 100213152
    download_obj = Downloader(
        dt_fr, dt_to, include_exp_ids=[100213152, 100213190],
        data_product='vi',
        include_file_name_patterns= [['pfa']],
        include_file_type_patterns=[['velocity']],
        exclude_file_type_patterns=['.*uncorrected.*', '.*power.*'],
        force_download=True,
        dry_run=False)


class Downloader(DownloaderBase):

    def __init__(
            self, dt_fr: datetime.datetime, dt_to: datetime,
            data_product=None,
            include_exp_name_patterns: list=None,
            exclude_exp_name_patterns: list=None,
            include_exp_ids: list = None,
            exclude_exp_ids: list=[100213840, ],
            include_file_name_patterns: list = None,
            exclude_file_name_patterns: list = None,
            include_file_type_patterns=None,
            exclude_file_type_patterns=None,
            data_file_root_dir: str = None,
            direct_download = True,
            force_download = False,
            dry_run: bool=False,
            madrigal_url: str = "http://cedar.openmadrigal.org/",
            user_fullname: str=madrigal.default_user_fullname,
            user_email: str=madrigal.default_user_email,
            user_affiliation: str=madrigal.default_user_affiliation):

        icodes = [61,]
        self.data_product = data_product if isinstance(data_product, str) else ''
        
        dt_fr = dttool.get_start_of_the_day(dt_fr)
        dt_to = dttool.get_end_of_the_day(dt_to)

        super().__init__(
            dt_fr=dt_fr, dt_to=dt_to, icodes=icodes,
            include_exp_name_patterns=include_exp_name_patterns, 
            exclude_exp_name_patterns=exclude_exp_name_patterns, 
            include_exp_ids=include_exp_ids,
            exclude_exp_ids=exclude_exp_ids,
            include_file_name_patterns=include_file_name_patterns, 
            exclude_file_name_patterns=exclude_file_name_patterns,
            include_file_type_patterns=include_file_type_patterns, 
            exclude_file_type_patterns=exclude_file_type_patterns,  
            data_file_root_dir=data_file_root_dir,
            force_download=force_download, direct_download=direct_download, dry_run=dry_run,
            madrigal_url=madrigal_url,
            user_fullname=user_fullname, user_email=user_email, user_affiliation=user_affiliation)

    def download(self, **kwargs):

        exps, database = self.get_exp_list(
            dt_fr=self.dt_fr,
            dt_to=self.dt_to,
            include_exp_name_patterns=self.include_exp_name_patterns,
            exclude_exp_name_patterns=self.exclude_exp_name_patterns,
            include_exp_ids=self.include_exp_ids,
            exclude_exp_ids=self.exclude_exp_ids,
            icodes=self.icodes,
            madrigal_url=self.madrigal_url,
            display=True)
        self.exp_list = list(exps)
        self.database = database

        exps, exps_error = self.get_online_file_list(
            exp_list=self.exp_list, database=database,
            include_file_name_patterns=self.include_file_name_patterns,
            exclude_file_name_patterns=self.exclude_file_name_patterns,
            include_file_type_patterns=self.include_file_type_patterns,
            exclude_file_type_patterns=self.exclude_file_type_patterns,
            display=True
        )
        self.exp_list_error = list(exps_error)

        file_paths = [] 
        for exp in exps:
            dt_fr_exp = datetime.datetime(
                exp.startyear, exp.startmonth, exp.startday, exp.starthour, exp.startmin, exp.startsec
            )
            dt_to_exp = datetime.datetime(
                exp.endyear, exp.endmonth, exp.endday, exp.endhour, exp.endmin, exp.endsec
            )
            for file in list(exp.files):

                file_path_remote = pathlib.Path(file.name)
                file_name_remote = file_path_remote.name

                res = re.search(r'/([\d]{2}[a-z]{3}[\d]{2})', file.doi)
                dtstr = res.groups()[0]
                dtstr = dtstr[0:2] + dtstr[2].upper() + dtstr[3:]
                thisday = datetime.datetime.strptime(dtstr, "%d%b%y")

                exp_name_patterns = re.findall('\w+', exp.name)
                if len(exp_name_patterns) > 5:
                    exp_name_patterns = exp_name_patterns[:5]
                exp_name = '_'.join(exp_name_patterns).lower()
                file_dir_local = self.data_file_root_dir / thisday.strftime("%Y") / \
                                ('FPISR_EID-' + str(exp.id) + '_'
                                 + dt_fr_exp.strftime("%Y%m%d%H%M%S") + '_'
                                 + dt_to_exp.strftime("%Y%m%d%H%M%S") + '_' + exp_name)
                file_dir_local.mkdir(parents=True, exist_ok=True)
                file_name_local = 'PFISR_' + thisday.strftime('%Y%m%d') + '_' + \
                                  self.data_product.replace(' ', '_') + \
                                  '.' + '.'.join(file_name_remote.split('.')[1:])
                file_path_local = file_dir_local / file_name_local
                if file_path_local.is_file():
                    mylog.simpleinfo.info("The file has been downloaded: {}.".format(file_path_local))
                    self.done = True
                    continue
                super().download(
                    file_path_remote=file.name, file_path_local=file_path_local,
                    file_format='hdf5')
                file_paths.append(file_path_local)
        return file_paths


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import requests
import bs4
import pathlib
import re
import netCDF4
import cftime

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.datahub.sources.wdc as wdc
from geospacelab.config import prf


class Downloader(object):

    def __init__(self, dt_fr,  dt_to, data_file_root_dir=None, user_email=wdc.default_user_email):

        self.dt_fr = dt_fr
        self.dt_to = dt_to

        self.user_email = user_email

        self.done = False
        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / 'WDC' / 'AE'
        else:
            self.data_file_root_dir = pathlib.Path(data_file_root_dir)

        self.url_base = "http://wdc.kugi.kyoto-u.ac.jp"

        self.download()

    def download(self):
        diff_months = dttool.get_diff_months(self.dt_fr, self.dt_to)
        dt0 = datetime.datetime(self.dt_fr.year, self.dt_fr.month, 1)
        r = requests.get(self.url_base + '/aeasy/')
        soup = bs4.BeautifulSoup(r.text, 'html.parser')
        form_tag = soup.find_all('form')
        r_method = form_tag[0].attrs['method']
        r_action_url = self.url_base + form_tag[0].attrs['action']
        for i in range(diff_months + 1):
            dt_fr = dttool.get_next_n_months(dt0, i)
            dt_to = dttool.get_next_n_months(dt0, i + 1) - datetime.timedelta(seconds=1)
            delta_seconds = (dt_to - dt_fr).total_seconds()

            file_name = 'WDC_AE_' + dt_fr.strftime('%Y%m') + '.nc'
            file_path = self.data_file_root_dir / '{:4d}'.format(dt_fr.year) / file_name
            if file_path.is_file():
                mylog.simpleinfo.info(
                    "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))
                self.done = True
                continue
            else:
                file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
            form_ae = {
                'Tens': str(int(dt_fr.year/10)),
                'Year': str(int(dt_fr.year - np.floor(dt_fr.year/10)*10)),
                'Month': '{:02d}'.format(dt_fr.month),
                'Day_Tens': '0',
                'Days':     '1',
                'Hour':    '00',
                'min':     '00',
                'Dur_Day_Tens': '{:02d}'.format(int(np.floor(np.ceil(delta_seconds/86400.)/10))),
                'Dur_Day': str(int(np.ceil(delta_seconds/86400.) - np.floor(np.ceil(delta_seconds/86400.)/10)*10)),
                'Dur_Hour': '00',
                "Dur_Min": '00',
                "Image Type": "GIF",
                "COLOR": "COLOR",
                "AE Sensitivity": "100",
                "ASY/SYM Sensitivity": "100",
                "Output": 'AE',
                "Out format": "IAGA2002",
                "Email": self.user_email,
            }

            if r_method.lower() == 'get':
                mylog.StreamLogger.info("Requesting data from WDC ...")
                r_file = requests.get(r_action_url, params=form_ae)

            if "No data for your request" in r_file.text or "DATE       TIME         DOY" not in r_file.text:
                mylog.StreamLogger.warning("No data for your request!")
                return

            with open(file_path.with_suffix('.dat'), 'w') as f:
                f.write(r_file.text)

            mylog.StreamLogger.info("Preparing to save the data in the netcdf format ...")
            self.save_to_netcdf(r_file.text, file_path)

    def save_to_netcdf(self, r_text, file_path):

        results = re.findall(
            r'^(\d+-\d+-\d+ \d+:\d+:\d+.\d+)\s*(\d+)\s*([+\-\d.]+)\s*([+\-\d.]+)\s*([+\-\d.]+)\s*([+\-\d.]+)',
            r_text,
            re.M
        )
        results = list(zip(*results))
        # time_array = np.array([(datetime.datetime.strptime(dtstr+'000', "%Y-%m-%d %H:%M:%S.%f")
        #                       - datetime.datetime(1970, 1, 1)) / datetime.timedelta(seconds=1)
        #                       for dtstr in results[0]])

        dts = [datetime.datetime.strptime(dtstr + '000', "%Y-%m-%d %H:%M:%S.%f") for dtstr in results[0]]
        time_array = np.array(cftime.date2num(dts, units='seconds since 1970-01-01 00:00:00.0'))
        print('From {} to {}.'.format(
            datetime.datetime.utcfromtimestamp(time_array[0]),
            datetime.datetime.utcfromtimestamp(time_array[-1]))
        )
        ae_array = np.array(results[2])
        ae_array.astype(np.float32)
        au_array = np.array(results[3])
        au_array.astype(np.float32)
        al_array = np.array(results[4])
        al_array.astype(np.float32)
        ao_array = np.array(results[5])
        ao_array.astype(np.float32)

        num_rows = len(results[0])

        fnc = netCDF4.Dataset(file_path, 'w')
        fnc.createDimension('UNIX_TIME', num_rows)

        fnc.title = "WDC AE indices"
        time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
        time.units = 'seconds since 1970-01-01 00:00:00.0'
        ae = fnc.createVariable('AE', np.float32, ('UNIX_TIME',))
        au = fnc.createVariable('AU', np.float32, ('UNIX_TIME',))
        al = fnc.createVariable('AL', np.float32, ('UNIX_TIME',))
        ao = fnc.createVariable('AO', np.float32, ('UNIX_TIME',))

        time[::] = time_array[::]
        ae[::] = ae_array[::]
        au[::] = au_array[::]
        al[::] = al_array[::]
        ao[::] = ao_array[::]
        # for i, res in enumerate(results):
        #     dt = datetime.datetime.strptime(res[0]+'000', "%Y-%m-%d %H:%M:%S.%f")
        #     time[i] = (dt - datetime.datetime(1970, 1, 1)) / datetime.timedelta(seconds=1)
        #     asy_d[i] = float(res[2])
        #     asy_h[i] = float(res[3])
        #     sym_d[i] = float(res[4])
        #     sym_h[i] = float(res[5])

        fnc.close()
        mylog.StreamLogger.info("The requested data has been downloaded and saved in the file {}.".format(file_path))
        self.done = True


if __name__ == "__main__":
    dt_fr1 = datetime.datetime(2000, 1, 14)
    dt_to1 = datetime.datetime(2000, 6, 16)
    Downloader(dt_fr1, dt_to1, user_email="lei.cai@oulu.fi")


import datetime
import pathlib
import copy

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf
from geospacelab.datahub.sources.cdaweb.dmsp.ssusi.downloader import Downloader as DownloaderBase


class Downloader(DownloaderBase):
    
    def __init__(
            self,
            dt_fr=None, dt_to=None,
            sat_id=None,
            orbit_id=None,
            direct_download=True,
            force_download=False,
            dry_run=False,
            root_dir_local = None,
        ):
         
        super().__init__(
            dt_fr, dt_to,
            sat_id=sat_id,
            orbit_id=orbit_id,
            product='SDR_DISK',
            root_dir_local=root_dir_local,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
        ) 


if __name__ == "__main__":
    downloader = Downloader(
        dt_fr = datetime.datetime(2011, 1, 6),
        dt_to = datetime.datetime(2011, 1, 6, 12),
        sat_id='F17',
        orbit_id='21523',
        force_download=True,
        dry_run=False,
    )


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import string
import pathlib
import matplotlib.pyplot as plt
import numpy as np
import datetime
import matplotlib as mpl
import matplotlib.dates as mdates
import cftime

from geospacelab.datahub import DataHub

import geospacelab.visualization.mpl.panels as mpl_panel
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pydatetime as dttool
from geospacelab.visualization.mpl.__base__ import DashboardBase
import geospacelab.visualization.mpl.panels as panels


class Dashboard(DataHub, DashboardBase):
    _default_layout_config = {
        'left': 0.15,
        'right': 0.78,
        'bottom': 0.15,
        'top': 0.88,
        'hspace': 0.1,
        'wspace': 0.1
    }

    def __init__(self, figure=None, figure_config=None, **kwargs):
        super(Dashboard, self).__init__(figure=figure, figure_config=figure_config, **kwargs)


class TSDashboard(Dashboard):
    _default_figure_config = {
        'figsize': (10, 8),
    }

    def __init__(
            self,
            figure=None, figure_config=None,
            time_gap=True,
            timeline_major='UT',
            timeline_extra_labels=None,
            timeline_same_format=False,
            **kwargs
    ):
        self.panel_layouts = []
        self.plot_styles = None
        self._panels_configs = {}

        figure_config = self._default_figure_config if figure_config is None else figure_config
        self.time_gap = time_gap
        self.timeline_major = timeline_major
        self.timeline_same_format = timeline_same_format
        self.timeline_extra_labels = timeline_extra_labels
        kwargs.update(visual='on')

        super(Dashboard, self).__init__(figure=figure, figure_config=figure_config, **kwargs)

        self.panels: dict[int: panels.TSPanel] = {}
        self._xlim = [self.dt_fr, self.dt_to]

    def set_layout(self, panel_layouts=None, panels_classes=None, plot_styles=None, row_height_scales=1,
                   left=None, right=None, bottom=None, top=None, hspace=None, add_panels=True, **kwargs):

        if left is None:
            left = self._default_layout_config['left']
        if right is None:
            right = self._default_layout_config['right']
        if bottom is None:
            bottom = self._default_layout_config['bottom']
        if top is None:
            top = self._default_layout_config['top']
        if hspace is None:
            hspace = self._default_layout_config['hspace']

        num_rows = len(panel_layouts)
        self.panel_layouts = panel_layouts
        if type(plot_styles) is not list:
            self.plot_styles = [None] * num_rows
        elif len(plot_styles) != num_rows:
            raise ValueError

        if panels_classes is None:
            panels_classes = [panels.TSPanel] * num_rows
        elif issubclass(panels_classes, panels.TSPanel):
            panels_classes = [panels_classes] * num_rows
        elif len(panels_classes) != num_rows:
            raise ValueError
        else:
            raise AttributeError

        if type(row_height_scales) is not list:
            if type(row_height_scales) == int:
                row_height_scales = [row_height_scales] * num_rows
            else:
                raise TypeError
        elif len(row_height_scales) != num_rows:
            raise ValueError

        hspace = hspace * row_height_scales[0]
        gs_num_rows = sum(row_height_scales)
        gs_num_cols = 1
        super().set_layout(num_rows=gs_num_rows, num_cols=gs_num_cols, left=left, right=right, bottom=bottom, top=top,
                           hspace=hspace, **kwargs)
        rec = 0
        for ind, height in enumerate(row_height_scales):
            self._panels_configs[ind] = {
                'row_ind': [rec, rec+height],
                'col_ind': [0, 1],
                'panel_class': panels_classes[ind],
            }
            rec = rec + height
        if add_panels:
            self.add_panels()

    def add_panels(self):
        bottom_panel = False
        for ind, panel_config in self._panels_configs.items():
            # share x
            if ind > 0:
                panel_config.update(sharex=self.panels[0]())
            if ind == len(self._panels_configs.keys())-1:
                bottom_panel = True
                panel_config.update(
                    timeline_extra_labels=self.timeline_extra_labels, timeline_same_format=self.timeline_same_format,
                )
            else:
                bottom_panel = False
            panel_config.update(bottom_panel=bottom_panel)
            self.add_panel(**panel_config)

    def draw(self, dt_fr=None, dt_to=None, auto_grid=True, time_res=None):
        if dt_fr is not None:
            self._xlim[0] = dt_fr
        if dt_to is not None:
            self._xlim[1] = dt_to

        if not isinstance(time_res,list):
            time_res = [time_res] * len(self.panels.keys())

        npanel = 0
        for ind, panel in self.panels.items():
            panel._xlim = self._xlim
            panel.time_res = time_res[ind]
            plot_layout = self.panel_layouts[npanel]
            plot_layout = self._validate_plot_layout(plot_layout)
            panel.draw(plot_layout)
            npanel = npanel + 1

        if auto_grid:
            self.add_grid(panel_id=0, visible=True, which='major', axis='both', lw=0.5, color='grey', alpha=0.3)
            self.add_grid(panel_id=0, visible=True, which='minor', axis='both', lw=0.3, color='grey', alpha=0.1)

    def add_grid(self, panel_id=0, visible=None, which='major', axis='both', **kwargs):
        if panel_id == 0:
            panels = self.panels.values()
        elif type(panel_id) is int:
            panels = [self.panels[int]]
        elif type(panel_id) is list:
            panels = [self.panels[i] for i in panel_id]
        else:
            raise NotImplementedError

        for p in panels:
            p.add_grid(ax=p(), visible=visible, which=which, axis=axis, **kwargs)

    def _validate_plot_layout(self, layout_in, level=0):
        from geospacelab.datahub import VariableModel
        if (level == 0) and (not isinstance(layout_in, list)):
            raise TypeError("The plot layout must be a list!")
        type_layout_in = type(layout_in)
        layout_out = []
        for ind, elem in enumerate(layout_in):
            if isinstance(elem, list):
                layout_out.append(self._validate_plot_layout(elem, level=level+1))
            elif issubclass(elem.__class__, VariableModel):
                var = elem
                layout_out.append(var)
            elif isinstance(elem, int):
                index = elem
                var = self.variables[index]
                layout_out.append(var)
            else:
                raise TypeError
        return type_layout_in(layout_out)

    def save_figure(self, file_dir=None, file_name=None, append_time=True, dpi=300, file_format='png', **kwargs):
        if file_dir is None:
            file_dir = pathlib.Path().cwd()
        else:
            file_dir = pathlib.Path(file_dir)
        if type(file_name) is not str:
            raise ValueError

        if append_time:
            dt_range_str = self.get_dt_range_str(style='filename')
            file_name = '_'.join([file_name, dt_range_str])

        file_name = file_name + '.' + file_format

        plt.savefig(file_dir / file_name, dpi=dpi, format=file_format, **kwargs)

    def add_title(self, x=0.5, y=1.08, title=None, **kwargs):
        append_time = kwargs.pop('append_time', True)
        kwargs.setdefault('fontsize', plt.rcParams['figure.titlesize'])
        kwargs.setdefault('fontweight', 'roman')
        if append_time:
            dt_range_str = self.get_dt_range_str(style='title')
            title = title + ', ' + dt_range_str
        title = title.replace(', , ', ', ')
        super().add_title(x=x, y=y, title=title, **kwargs)

    def get_dt_range_str(self, style='title'):
        dt_fr = self._xlim[0]
        dt_to = self._xlim[1]
        if style == 'title':
            diff_days = dttool.get_diff_days(dt1=dt_fr, dt2=dt_to)
            if diff_days == 0:
                fmt1 = "%Y-%m-%dT%H:%M:%S"
                fmt2 = "%H:%M:%S"
            else:
                fmt1 = "%Y-%m-%dT%H:%M:%S"
                fmt2 = fmt1
            dt_range_str = dt_fr.strftime(fmt1) + ' - ' + dt_to.strftime(fmt2)
        elif style == 'filename':
            diff_days = dttool.get_diff_days(dt1=dt_fr, dt2=dt_to)
            if diff_days == 0:
                fmt1 = "%Y%m%d-%H%M%S"
                fmt2 = "%H%M%S"
            else:
                fmt1 = "%Y%m%d-%H%M%S"
                fmt2 = fmt1
            dt_range_str = dt_fr.strftime(fmt1) + '-' + dt_to.strftime(fmt2)
        return dt_range_str

    def add_vertical_line(self, dt_in, panel_index=-1,
                          label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):
        if type(dt_in) not in [datetime.datetime, cftime.real_datetime]:
            return

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('linewidth', 2)
        kwargs.setdefault('linestyle', '--')
        kwargs.setdefault('color', 'k')
        kwargs.setdefault('alpha', 0.8)

        if panel_index == -1:
            if 'major' not in self.extra_axes.keys():
                ax = self.add_major_axes()
            else:
                ax = self.extra_axes['major']
        else:
            ax = self.panels[panel_index].axes['major']

        xlim = self.panels[0].axes['major'].get_xlim()
        ylim = ax.get_ylim()
        diff_xlim = xlim[1] - xlim[0]
        diff_ylim = ylim[1] - ylim[0]
        x = mdates.date2num(dt_in)
        x = (x - xlim[0]) / diff_xlim

        y0 = 0 - bottom_extend
        y1 = 1. + top_extend
        line = mpl.lines.Line2D([x, x], [y0, y1], transform=ax.transAxes, **kwargs)
        line.set_clip_on(False)
        ax.add_line(line)

        if type(label) is str:
            if label_position == 'top':
                y = y1
                text_config.setdefault('va', 'bottom')
            elif label_position == 'bottom':
                y = y0
                text_config.setdefault('va', 'top')
            else:
                x = label_position[0]
                y = label_position[1]
                text_config.setdefault('va', 'center')
            text_config.setdefault('ha', 'center')
            text_config.setdefault('fontsize', plt.rcParams['axes.labelsize'])
            text_config.setdefault('fontweight', 'medium')
            text_config.setdefault('clip_on', False)
            ax.text(x, y, label, transform=ax.transAxes, **text_config)

    def add_horizontal_line(self, ys, panel_index=-1, *,
                          label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('linewidth', 1)
        kwargs.setdefault('linestyle', '--')
        kwargs.setdefault('color', 'k')
        kwargs.setdefault('alpha', 0.8)

        if panel_index == -1:
            ax_list = [p.axes['major'] for p in self.panels]
        else:
            ax_list = [self.panels[panel_index].axes['major']]

        if type(ys) is not list:
            ys = [ys] * len(ax_list)
        for ind, ax in enumerate(ax_list):
            xx = ax.get_xlim()
            yy = [ys[ind], ys[ind]]
            ax.plot(xx, yy, **kwargs)


    def search_UTs(self, search_step=1/86400, **kwargs) -> list:
        from scipy.interpolate import interp1d
        from scipy.signal import argrelmin
        import matplotlib.dates as mpl_dates
        panel = list(self.panels.values())[-1]
        var_for_config = panel._var_for_config
        x_depend = var_for_config.get_depend(axis=0, retrieve_data=True)
        x0 = np.array(mpl_dates.date2num(x_depend['UT'])).flatten()
        x_fr = x0[0]
        x_to = x0[-1]
        x1 = np.arange(x_fr, x_to, search_step)
        x1_0 = x1
        ind_0 = np.array(range(x1.size))
        for ind, (label, value) in enumerate(kwargs.items()):
            if label in x_depend.keys():
                y0 = x_depend[label].flatten()
            elif label in var_for_config.dataset.keys():
                y0 = var_for_config.dataset[label].value
            else:
                raise KeyError
            if ind == 0:
                if 'MLT' in label.upper():
                    y0_sin = np.sin(y0 / 24. * 2 * np.pi)
                    y0_cos = np.cos(y0 / 24. * 2 * np.pi)
                    itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                    itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                    y0_sin_i = itpf_sin(x1)
                    y0_cos_i = itpf_cos(x1)
                    rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                    rad = np.where((rad >= 0), rad, rad + 2 * np.pi)
                    y1 = rad / 2. / np.pi * 24.
                elif 'LON' in label.upper():
                    y0_sin = np.sin(y0 * np.pi / 180.)
                    y0_cos = np.cos(y0 * np.pi / 180.)
                    itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                    itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                    y0_sin_i = itpf_sin(x1)
                    y0_cos_i = itpf_cos(x1)
                    rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                    y1 = rad * 180. / np.pi
                    if value is list:
                        value = [(v + 180.) % 360 - 180 for v in value]
                    else:
                        value = (value + 180.) % 360 - 180
                else:
                    itpf = interp1d(x0, y0, bounds_error=False, fill_value='extrapolate')
                    y1 = itpf(x1)
            else:
                y1 = y0
            if type(value) in (int, float):
                ind_1 = argrelmin(np.abs(y1 - value))
            elif type(value) is list:
                ind_1 = np.where((y1 > value[0]) & (y1 < value[1]))[0]
            else:
                raise NotImplementedError

            x1 = x1[ind_1]
            y1 = y1[ind_1]

        results = []
        for xx in x1:
            ind_1 = argrelmin(np.abs(x0 - xx))
            results.extend(mpl_dates.num2date(x0[ind_1]))

        return results

    def add_shading(self, dt_fr, dt_to, panel_index=-1,
                    label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('edgecolor', 'none')
        kwargs.setdefault('facecolor', 'yellow')
        kwargs.setdefault('alpha', 0.4)

        if panel_index == -1:
            if 'major' not in self.extra_axes.keys():
                ax = self.add_major_axes()
            else:
                ax = self.extra_axes['major']
        else:
            ax = self.panels[panel_index].axes['major']

        xlim = self.panels[0].axes['major'].get_xlim()
        ylim = ax.get_ylim()
        diff_xlim = xlim[1] - xlim[0]
        diff_ylim = ylim[1] - ylim[0]
        x0 = mdates.date2num(dt_fr)
        x0 = (x0 - xlim[0]) / diff_xlim
        x1 = mdates.date2num(dt_to)
        x1 = (x1 - xlim[0]) / diff_xlim

        y0 = 0 - bottom_extend
        y1 = 1. + top_extend

        rectangle = mpl.patches.Rectangle((x0, y0), x1-x0, y1-y0, transform=ax.transAxes, **kwargs)
        rectangle.set_clip_on(False)
        ax.add_patch(rectangle)

        if type(label) is str:
            x = (x0 + x1) / 2
            if label_position == 'top':
                y = y1
                text_config.setdefault('va', 'bottom')
            elif label_position == 'bottom':
                y = y0
                text_config.setdefault('va', 'top')
            else:
                x = label_position[0]
                y = label_position[1]
                text_config.setdefault('va', 'center')
            text_config.setdefault('ha', 'center')
            text_config.setdefault('fontsize', plt.rcParams['axes.labelsize'])
            text_config.setdefault('fontweight', 'medium')
            text_config.setdefault('clip_on', False)
            ax.text(x, y, label, transform=ax.transAxes, **text_config)

    def add_top_bar(self, dt_fr, dt_to, panel_index=-1, bottom=0, top=0.02, facecolor='orange', **kwargs):
        bottom_extend = -1. - bottom
        top_extend = top
        kwargs.setdefault('alpha', 1.)
        self.add_shading(dt_fr, dt_to, panel_index=panel_index, bottom_extend=bottom_extend, top_extend=top_extend, facecolor=facecolor, **kwargs)




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
from geospacelab.config import prf

from geospacelab.datahub.sources.tud.downloader import TUDownloader as DownloaderBase   


class Downloader(DownloaderBase):
    
    def __init__(self, 
        dt_fr, dt_to, 
        mission='GOCE', 
        sat_id='',
        version='v01',  
        product='DNS-WND-ACC',
        direct_download=True,
        force_download=False,
        dry_run=False,
        ):
        self.mission = mission
        self.version = version
        
        super().__init__(
            dt_fr, dt_to, 
            mission=mission,
            sat_id=sat_id, 
            version=version,
            product=product,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run
        )
        


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import datetime

import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, InstrumentModel, ProductModel
from geospacelab.datahub.sources.cdaweb import cdaweb_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.sources.cdaweb.dmsp.ssusi.edr_aur.loader import Loader as default_Loader
from geospacelab.datahub.sources.cdaweb.dmsp.ssusi.edr_aur.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.cdaweb.dmsp.ssusi.edr_aur.variable_config as var_config


default_dataset_attrs = {
    'database': cdaweb_database,
    'facility': 'DMSP',
    'instrument': 'SSUSI',
    'product': 'EDR_AUR',
    'data_file_ext': 'nc',
    'data_root_dir': prf.datahub_data_root_dir / 'CDAWeb' / 'DMSP' / 'SSUSI' / 'EDR_AUR',
    'allow_load': True,
    'allow_download': True,
    'data_search_recursive': False,
    'label_fields': ['database', 'facility', 'instrument', 'product'],
    'time_clip': False,
}

default_variable_names = [
    'DATETIME', 'STARTING_TIME', 'STOPPING_TIME',
    'ORBIT_ID',
    'GRID_MLAT', 'GRID_MLON', 'GRID_MLT', 'GRID_UT',
    'GRID_AUR_1216', 'GRID_AUR_1304', 'GRID_AUR_1356', 'GRID_AUR_LBHS', 'GRID_AUR_LBHL',
]

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', 'CDAWeb')
        self.facility = kwargs.pop('facility', 'DMSP')
        self.instrument = kwargs.pop('instrument', 'SSUSI')
        self.product = kwargs.pop('product', 'EDR_AUR')
        self.allow_download = kwargs.pop('allow_download', True)
        self.force_download = kwargs.pop('force_download', False)
        self.download_dry_run = kwargs.pop('download_dry_run', False)

        self.sat_id = kwargs.pop('sat_id', '')
        self.orbit_id = kwargs.pop('orbit_id', None)
        self.pole = kwargs.pop('pole', '')

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )
        if self.orbit_id is None:
            self.orbit_id = []
        for file_path in self.data_file_paths:
            try:
                load_obj = self.loader(file_path, file_type=self.product.lower(), pole=self.pole)
            except Exception as e:
                mylog.StreamLogger.warning(("Cannot load the data file: {}".format(file_path)))
                print(e)
            for var_name in self._variables.keys():
                if var_name == 'EMISSION_SPECTRA':
                    self._variables[var_name].value = load_obj.variables[var_name]
                    continue
                if var_name in ['DATETIME', 'STARTING_TIME', 'STOPPING_TIME']:
                    value = np.array([load_obj.variables[var_name]])[np.newaxis, :]
                elif var_name == 'ORBIT_ID':
                    value = np.array([load_obj.metadata['ORBIT_ID']], dtype=str) 
                else:
                    value = np.empty((1, ), dtype=object)
                    value[0] = load_obj.variables[var_name]
                    # value = load_obj.variables[var_name][np.newaxis, ::]
                self._variables[var_name].join(value)

        self.orbit_id = self['ORBIT_ID'].value
        if len(self.orbit_id):
            self.orbit_id = self.orbit_id[0]
        if self.time_clip:
            self.time_filter_by_range()
        
        del load_obj

    def get_time_ind(self, ut, time_res=20*60, var_datetime_name='DATETIME', edge_cutoff=False, **kwargs):
        ind = super().get_time_ind(ut, time_res=time_res, var_datetime_name=var_datetime_name, edge_cutoff=edge_cutoff, **kwargs)
        return ind

    def search_data_files(self, **kwargs):
        dt_fr_1 = self.dt_fr - datetime.timedelta(hours=3)
        dt_to_1 = self.dt_to + datetime.timedelta(hours=3)

        diff_days = dttool.get_diff_days(dt_fr_1, dt_to_1)
        
        dt0 = dttool.get_start_of_the_day(dt_fr_1)
        for i in range(diff_days + 1):
            thisday = dt0 + datetime.timedelta(days=i)
            initial_file_dir = kwargs.pop('initial_file_dir', None)
            if initial_file_dir is None:
                initial_file_dir = self.data_root_dir / self.sat_id.upper() / str(thisday.year) /thisday.strftime("%Y%m%d")

            file_patterns = [
                'dmsp' + self.sat_id.lower(),
                'edr-aurora',
                thisday.strftime("%Y%j") + 'T',
            ]

            if self.orbit_id is not None:
                file_patterns.extend(['REV', self.orbit_id])

            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            if self.orbit_id is not None:
                multiple_files = False
            else:
                fp_log = initial_file_dir / (self.product.upper() + '.full.log')
                if not fp_log.is_file():
                    self.download_data(dt_fr=thisday, dt_to=thisday)
                multiple_files = True
            
            done = super().search_data_files(
                initial_file_dir=initial_file_dir,
                search_pattern=search_pattern,
                allow_multiple_files=multiple_files,
            )
            if done and self.orbit_id is not None:
                return True

            # Validate file paths

            if (not done and self.allow_download) or (self.force_download):
                done = self.download_data(dt_fr=thisday, dt_to=thisday)
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir,
                        search_pattern=search_pattern,
                        allow_multiple_files=multiple_files
                    )

        return done

    def download_data(self, dt_fr=None, dt_to=None):
        if dt_fr is None:
            dt_fr = self.dt_fr
        if dt_to is None:
            dt_to = self.dt_to
        download_obj = self.downloader(
            dt_fr, dt_to,
            orbit_id=self.orbit_id, sat_id=self.sat_id,
            root_dir_local=self.data_root_dir,
            force_download=self.force_download,
            dry_run=self.download_dry_run,
        )
        return any(download_obj.done)

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        if isinstance(value, str):
            self._instrument = InstrumentModel(value)
        elif issubclass(value.__class__, InstrumentModel):
            self._instrument = value
        else:
            raise TypeError



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import cftime
import datetime


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, file_type='TEC-MAP', load_data=True):
        self.file_path = file_path
        self.file_type = file_type
        self.variables = {}
        self.done = False
        if load_data:
            self.load()

    def load(self):
        if self.file_type == "TEC-MAP":
            self.load_tec_map()
        else:
            raise NotImplemented

    def load_tec_map(self):
        variables = {}
        with h5py.File(self.file_path, 'r') as fh5:
            tec = fh5['Data']['Array Layout']['2D Parameters']['tec'][:]
            variables['TEC_MAP'] = np.transpose(tec, (2, 0, 1))

            t = fh5['Data']['Array Layout']['timestamps'][:]
            t = cftime.num2date(
                t,
                units="seconds since 1970-01-01 00:00:00.0",
                only_use_cftime_datetimes=False,
                only_use_python_datetimes=True,
            )
            variables['DATETIME'] = np.reshape(t, (t.size, 1))

            glon =fh5['Data']['Array Layout']['glon'][:]
            glat = fh5['Data']['Array Layout']['gdlat'][:]
            variables['GEO_LAT'] = np.tile(glat.reshape(glat.size, 1), (1, glon.size))
            variables['GEO_LON'] = np.tile(glon.reshape(1, glon.size), (glat.size, 1))
        self.variables = variables



if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path("/Users/lcai/Downloads/gps200102g.002.hdf5")
    Loader(file_path=fp)
    # import pathlib
    # import geospacelab.datahub.sources.madrigal as madrigal
    # file_path = pathlib.Path("/Users/lcai/Downloads/gps200102g.002.hdf5")
    # madrigal.utilities.show_hdf5_structure(filename=file_path.name, filepath= file_path.parent.resolve())
    # madrigal.utilities.show_hdf5_metadata(filename=file_path.name, filepath=file_path.parent.resolve())
    # with h5py.File(file_path, 'r') as fh5:
    #
    #     print(fh5['Data']['Array Layout']['2D Parameters']['tec'][:])  # shape (180, 360, 288)
    #     print(fh5['Data']['Array Layout']['timestamps'][:])  # shape (288) unix time
    #     print(fh5['Data']['Array Layout']['glon'][:])  # gdlat -90. - 89. glon -180. 79.


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import requests
import bs4
import pathlib
import re
import zipfile
import ftplib
from contextlib import closing

from geospacelab.datahub.__dataset_base__ import DownloaderBase
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf

from geospacelab.datahub.sources.cdaweb.downloader import DownloaderFromHTTPBase as DownloaderBase


class TUDownloader(DownloaderBase):
    
    root_dir_remote = 'data/data'
    
    def __init__(
        self, 
        dt_fr, dt_to, 
        mission='SWARM',
        sat_id=None,
        product='DNS-POD',
        version='v02',
        root_dir_local=None, 
        direct_download=True,
        force_download=False, dry_run=False, 
        **kwargs):
        
        base_url = 'https://thermosphere.tudelft.nl'
        if root_dir_local is None:
            root_dir_local = prf.datahub_data_root_dir / 'TUD'
            
        self.mission = mission
        self.sat_id = sat_id
        self.product = product
        self.version = version
        
        self._validate_mission_name()
        self._validate_sat_id()
        self._validate_product()
        self._validate_version()
        super().__init__(
            dt_fr, dt_to, 
            base_url=base_url, 
            root_dir_local=root_dir_local, 
            direct_download=direct_download, 
            force_download=force_download, 
            dry_run=dry_run, **kwargs)
    
    def search_from_http(self, subdirs = None, file_name_patterns = None, **kwargs):
        
        diff_months = dttool.get_diff_months(self.dt_fr, self.dt_to)
        file_paths_remote = []
        for nm in range(diff_months + 1):
            this_month = dttool.get_next_n_months(self.dt_fr, nm)
            
            subdirs = [version_dict[self.version[:3]]]
            subdirs.append(self.mission + '_data')

            file_name_patterns = [
                self.mission[0] + self.sat_id,
                self.product.replace('-', '_'),
                this_month.strftime('%Y'),
                this_month.strftime('%m'),
                self.version
            ]
            paths = super().search_from_http(subdirs, file_name_patterns, **kwargs)
            if len(paths) > 1:
                mylog.StreamLogger.warning(
                    f"Multiple versions found for {this_month.strftime('%Y-%m')}. All versions will be downloaded.")
            file_paths_remote.extend(paths)
        return file_paths_remote
    
    def save_files_from_http(self, file_paths_local=None, root_dir_remote=None):
        
        file_paths_local = []
        for fp_remote in self.file_paths_remote:
            fn_remote = fp_remote.split('/')[-1]
            
            rc = re.compile(r'_(v[\w]+)')
            rm = rc.search(fn_remote)
            if rm is None:
                raise ValueError(f"Cannot extract version info from file name {fn_remote}.")
            version_in_fn = rm.groups()[0]
            
            replacement = '/'.join([s.strip('/') for s in [
                self.base_url, 
                self.root_dir_remote,
                version_dict[version_in_fn[:3]], 
                self.mission + '_data'
            ]]) + '/'
            fp_local = self.root_dir_local / self.mission.upper() / self.product.upper() / version_in_fn
            for p in fp_remote.replace(replacement, '').split('/'):
                fp_local = fp_local / p
            file_paths_local.append(fp_local)
        super().save_files_from_http(file_paths_local, root_dir_remote)  
        
        for i, (done, file_path) in enumerate(zip(self.done, self.file_paths_local)):
            if not done:
                continue
            else:
                if not file_path.exists():
                    continue
                mylog.simpleinfo.info(f"Uncompressing the file: {file_path} ...")
                with zipfile.ZipFile(file_path, 'r') as zip_ref:
                    zip_ref.extractall(file_path.parent.resolve())
                    file_path.unlink()
                    mylog.simpleinfo.info("Done. The zip file has been removed.")
                    self.file_paths_local[i] = file_path.with_suffix('.txt') # Update to unzipped file path
        return 
    
    def _to_download(self, file_path, with_suffix='.txt'):
        return super()._to_download(file_path, with_suffix=with_suffix)
    
    def _validate_mission_name(self):
        if self.mission.lower() == 'swarm':
            self.mission = 'Swarm'
        elif self.mission.lower() in ['champ', 'grace', 'goce', 'grace-fo']:
            self.mission = self.mission.upper()
        else:
            raise NotImplementedError(f"Mission {self.mission} is not supported.")
        
    def _validate_product(self):
        if self.mission == 'Swarm' and self.sat_id in ['A', 'B', 'C']:
            valid_products = ['DNS-POD', 'DNS-ACC']
            assert self.product in valid_products, f"Invalid product {self.product}. Valid products are {valid_products}."
        elif self.mission in ['CHAMP', 'GRACE', 'GOCE', 'GRACE-FO']:
            valid_products = ['DNS-ACC', 'WND-ACC']
            assert self.product in valid_products, f"Invalid product {self.product}. Valid products are {valid_products}."
        else:
            raise NotImplementedError(f"Mission {self.mission} with sat_id {self.sat_id} is not supported.")
        return
            
        
    def _validate_sat_id(self):
        
        if self.mission.upper() == 'SWARM':
            valid_sat_ids = ['A', 'B', 'C']
            assert self.sat_id in valid_sat_ids, f"Invalid sat_id {self.sat_id}. Valid sat_ids are {valid_sat_ids}."
        elif self.mission.upper() == 'CHAMP':
            valid_sat_ids = ['']
            assert self.sat_id in valid_sat_ids, f"Invalid sat_id {self.sat_id}. Valid sat_ids are {valid_sat_ids}."
        elif self.mission.upper() == 'GRACE':
            valid_sat_ids = ['A', 'B']
            assert self.sat_id in valid_sat_ids, f"Invalid sat_id {self.sat_id}. Valid sat_ids are {valid_sat_ids}."
        elif self.mission.upper() == 'GOCE':
            valid_sat_ids = ['']
            assert self.sat_id in valid_sat_ids, f"Invalid sat_id {self.sat_id}. Valid sat_ids are {valid_sat_ids}."
        elif self.mission.upper() == 'GRACE-FO':
            valid_sat_ids = ['FO1', '1', 'C']
            assert self.sat_id in valid_sat_ids, f"Invalid sat_id {self.sat_id}. Valid sat_ids are {valid_sat_ids}."
            self.sat_id = 'C'  # Standardize to 'C' for GRACE-FO
        else:
            raise NotImplementedError(f"Mission {self.mission} is not supported.")
        return

    def _validate_version(self):
        if self.mission.upper() in ['SWARM', 'CHAMP', 'GOCE', 'GRACE']:
            valid_versions = ['v01', 'v02']
            assert self.version in valid_versions, f"Invalid version {self.version}. Valid versions are {valid_versions}."
        elif self.mission.upper() == 'GRACE-FO':
            valid_versions = ['v02']
            assert self.version in valid_versions, f"Invalid version {self.version}. Valid versions are {valid_versions}."
        else:
            raise NotImplementedError(f"Mission {self.mission} is not supported.")
        return

version_dict = {
    'v01': 'version_01',
    'v02': 'version_02',
}


# class Downloader(DownloaderBase):
#     """
#     Base downloader for downloading the SWARM data files from ftp://swarm-diss.eo.esa.int/

#     :param ftp_host: the FTP host address
#     :type ftp_host: str
#     :param ftp_port: the FTP port [21].
#     :type ftp_port: int
#     :param ftp_data_dir: the directory in the FTP that stores the data.
#     :type ftp_data_dir: str
#     """

#     def __init__(self,
#                  dt_fr, dt_to,
#                  data_file_root_dir=None, ftp_data_dir=None, force=True, direct_download=True, file_name_patterns=[],
#                  **kwargs):
#         self.ftp_host = "thermosphere.tudelft.nl"
#         self.ftp_port = 21
#         if ftp_data_dir is None:
#             raise ValueError

#         self.ftp_data_dir = ftp_data_dir

#         self.file_name_patterns = file_name_patterns

#         super(Downloader, self).__init__(
#             dt_fr, dt_to, data_file_root_dir=data_file_root_dir, force=force, direct_download=direct_download, **kwargs
#         )


#     def download_from_http(self, **kwargs):
        
#         return
    
    
#     def download_from_ftp(self, **kwargs):
#         """[DEPRECATED] Download data files from FTP server.

#         Returns:
#             _type_: _description_
#         """
#         done = False
#         try:
#             ftp = ftplib.FTP()
#             ftp.connect(self.ftp_host, self.ftp_port, 30)  # 30 timeout
#             ftp.login()
#             ftp.cwd(self.ftp_data_dir)
#             file_list = ftp.nlst()

#             file_names = self.search_files(file_list=file_list, file_name_patterns=self.file_name_patterns)
#             file_dir_root = self.data_file_root_dir
#             for ind_f, file_name in enumerate(file_names):

#                 file_path = file_dir_root / file_name

#                 if file_path.is_file():
#                     mylog.simpleinfo.info(
#                         "The file {} exists in the directory {}.".format(
#                             file_path.name, file_path.parent.resolve()
#                         )
#                     )
#                     if not self.force:
#                         done = True
#                         continue
#                 else:
#                     file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
#                 mylog.simpleinfo.info(
#                     f"Downloading the file {file_name} from the FTP ..."
#                 )
#                 try:
#                     with open(file_path, 'w+b') as f:
#                         done = False
#                         res = ftp.retrbinary('RETR ' + file_name, f.write)
#                         print(res)
#                         if not res.startswith('226'):
#                             mylog.StreamLogger.warning('Downloaded of file {0} is not compile.'.format(file_name))
#                             pathlib.Path.unlink(file_path)
#                             done = False
#                             return done
#                         mylog.simpleinfo.info("Done.")
#                 except:
#                     pathlib.Path.unlink(file_path)
#                     mylog.StreamLogger.warning('Downloaded of file {0} is not compile.'.format(file_name))
#                     return False
#                 mylog.simpleinfo.info("Uncompressing the file ...")
#                 with zipfile.ZipFile(file_path, 'r') as zip_ref:
#                     zip_ref.extractall(file_path.parent.resolve())
#                     file_path.unlink()
#                 mylog.simpleinfo.info("Done. The zip file has been removed.")

#             done = True
#             ftp.quit()
#         except Exception as err:
#             print(err)
#             print('Error during download from FTP')
#             done = False
#         return done

#     def search_files(self, file_list=None, file_name_patterns=None):
#         def extract_timeline(files):
            
#             nf = len(files)
#             dt_list = []
#             for ind, fn in enumerate(files):
#                 dt_regex = re.compile(r'_(\d{4}_\d{2})_')
#                 rm = dt_regex.findall(fn)
#                 if not list(rm):
#                     continue
#                 dt_list.append(datetime.datetime.strptime(rm[0] + '_01', '%Y_%m_%d'))
#             dt_list = np.array(dt_list, dtype=datetime.datetime)

#             return dt_list

#         if file_name_patterns is None:
#             file_name_patterns = []

#         if list(file_name_patterns):
#             search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
#             fn_regex = re.compile(search_pattern)
#             file_list = list(filter(fn_regex.match, file_list))

#         dt_list = extract_timeline(file_list)

#         dt_fr = datetime.datetime(self.dt_fr.year, self.dt_fr.month, 1)
#         dt_to = datetime.datetime(self.dt_to.year, self.dt_to.month, 1)
#         ind_dt = np.where((dt_list >= dt_fr) & (dt_list <= dt_to))[0]
#         if not list(ind_dt):
#             raise FileExistsError
#         file_list = [file_list[ii] for ii in ind_dt]

#         return file_list








# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import netCDF4
import datetime
import numpy as np
import re
import cdflib

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as pybasic


class Loader(object):

    def __init__(self, file_path, file_type='sdr-disk', pole='N', pp_type='AURORAL'):

        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type
        self.pole = pole
        self.pp_type = pp_type

        self.load_data()

    def load_data(self):
        dataset = netCDF4.Dataset(self.file_path)
        variables = {}
        metadata = {}

        # get the version of the file
        variables['FILE_VERSION'] = re.findall(r"(v[\w\d.]+).nc", dataset.FILENAME)[0]
        variables['DATA_PRODUCT_VERSION'] = str(dataset.DATA_PRODUCT_VERSION)
        variables['SOFTWARE_VERSION_NUMBER'] = str(dataset.SOFTWARE_VERSION_NUMBER)
        variables['CALIBRATION_PERIOD_VERSION'] = dataset.CALIBRATION_PERIOD_VERSION
        variables['EMISSION_SPECTRA'] = ['1216', '1304', '1356', 'LBHS', 'LBHL']

        if self.pole == 'N':
            pole = self.pole
            pole_str = 'NORTH'
        elif self.pole == 'S':
            pole = self.pole
            pole_str = 'SOUTH'
        else:
            raise ValueError

        if self.pp_type in ['AURORAL', 'DAY_AURORAL', 'DAY_AUR']:
            pp_key_1 = 'DAY'
            pp_key_2 = 'AURORAL'
        else:
            pp_key_1 = self.pp_type
            pp_key_2 = ''

        # Global attributes
        starting_time = datetime.datetime.strptime(dataset.STARTING_TIME, "%Y%j%H%M%S")
        variables['STARTING_TIME'] = starting_time
        stopping_time = datetime.datetime.strptime(dataset.STOPPING_TIME, "%Y%j%H%M%S")
        variables['STOPPING_TIME'] = stopping_time
        dt0 = dttool.get_start_of_the_day(starting_time)

        # Time and position
        sectime = pybasic.str_join('TIME', pp_key_1, pp_key_2, separator='_', uppercase=True)
        sectime = np.array(dataset.variables[sectime])
        cdfepoch = pybasic.str_join('TIME', 'EPOCH', pp_key_1, pp_key_2, separator='_', uppercase=True)
        dts = cdflib.cdfepoch.unixtime(np.array(dataset.variables[cdfepoch]))
        dts = dttool.convert_unix_time_to_datetime_cftime(dts)

        var_name_nc = pybasic.str_join('ORBIT', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['SC_ORBIT_ID'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = pybasic.str_join('LATITUDE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        geo_lat = np.array(dataset.variables[var_name_nc])

        ind_full = np.arange(0, geo_lat.shape[0])
        if self.pole in ['N', 'S']:
            ind_S = np.where(geo_lat < -5.)[0]
            if list(ind_S):
                ind_N = ind_full[0:ind_S[0]]
            else:
                ind_N = np.where(geo_lat > 5.)[0]
            if (self.pole == 'N' and not list(ind_N)) or (self.pole == 'S' and not list(ind_S)):
                raise ValueError("No data available!")

            if self.pole == 'N':
                ind_along = ind_N
            else:
                ind_along = ind_S
        elif self.pole == 'both':
            ind_along = ind_full
        else:
            raise NotImplementedError

        variables['SC_DATETIME'] = dts[ind_along]

        variables['SC_GEO_LAT'] = geo_lat[ind_along]

        if self.pole in ['N', 'S']:
            ind_center = np.where(np.abs(variables['SC_GEO_LAT']) == np.max(np.abs(variables['SC_GEO_LAT'])))[0][0]
            variables['DATETIME'] = variables['SC_DATETIME'][ind_center]
        else:
            variables['DATETIME'] = starting_time + (stopping_time - starting_time) / 2

        var_name_nc = pybasic.str_join('LONGITUDE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['SC_GEO_LON'] = np.array(dataset.variables[var_name_nc])[ind_along]

        var_name_nc = pybasic.str_join('ALTITUDE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['SC_GEO_ALT'] = np.array(dataset.variables[var_name_nc])[ind_along]

        var_name_nc = pybasic.str_join('PIERCEPOINT', pp_key_1, 'LATITUDE', pp_key_2, separator='_', uppercase=True)
        variables['DISK_GEO_LAT'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        var_name_nc = pybasic.str_join('PIERCEPOINT', pp_key_1, 'LONGITUDE', pp_key_2, separator='_', uppercase=True)
        variables['DISK_GEO_LON'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        var_name_nc = pybasic.str_join('PIERCEPOINT', pp_key_1, 'ALTITUDE', pp_key_2, separator='_', uppercase=True)
        variables['DISK_GEO_ALT'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = pybasic.str_join('PIERCEPOINT', pp_key_1, 'SZA', pp_key_2, separator='_', uppercase=True)
        variables['DISK_SZA'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        var_name_nc = pybasic.str_join('IN_SAA', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['DISK_SAA'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        var_name_nc = pybasic.str_join('ACROSSPIXELSIZE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['ACROSS_PIXEL_SIZE'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = pybasic.str_join('ALONGPIXELSIZE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['ALONG_PIXEL_SIZE'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = pybasic.str_join('EFFECTIVELOOKANGLE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['EFFECTIVE_LOOK_ANGLE'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        # Re-binned data
        # var_name_nc = pybasic.str_join('DISKCOUNTSDATA', pp_key_1, pp_key_2, separator='_', uppercase=True)
        # variables['DISK_COUNTS_DATA'] = np.array(dataset.variables[var_name_nc])

        # var_name_nc = pybasic.str_join('DISKDECOMP_UNCERTAINTY', pp_key_1, pp_key_2, separator='_', uppercase=True)
        # variables['DISK_DECOMP_UNCERTAINTY'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = pybasic.str_join('EXPOSURE', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['EXPOSURE'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        var_name_nc = pybasic.str_join('SAA_COUNT', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['SAA_COUNT'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        # Calibration parameters
        var_name_nc = 'DARK_COUNT_CORRECTION'
        variables['DARK_COUNT_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = 'SCATTER_LIGHT_1216_CORRECTION'
        variables['SCATTER_LIGHT_1216_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = 'SCATTER_LIGHT_1304_CORRECTION'
        variables['SCATTER_LIGHT_1304_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = 'OVERLAP_1304_1356_CORRECTION'
        variables['OVERLAP_1304_1356_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = 'LONGWAVE_SCATTER_CORRECTION'
        variables['LONG_WAVE_SCATTER_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        var_name_nc = 'RED_LEAK_CORRECTION'
        variables['RED_LEAK_CORRECTION'] = np.array(dataset.variables[var_name_nc])

        # Calibrated data
        var_name_nc = pybasic.str_join('DQI', pp_key_1, pp_key_2, separator='_', uppercase=True)
        variables['DQI'] = np.array(dataset.variables[var_name_nc]).T[ind_along, :]

        dict_vn_nc = {
            'DISK_COUNTS':  pybasic.str_join('DISKCOUNTSDATA', pp_key_1, pp_key_2, separator='_', uppercase=True),
            'DISK_R': pybasic.str_join('DISK', 'INTENSITY', pp_key_1, pp_key_2, separator='_', uppercase=True),
            'DISK_R_RECT': pybasic.str_join('DISK', 'RECTIFIED', 'INTENSITY', pp_key_1, pp_key_2, separator='_', uppercase=True),
            'DISK_DECOMP_ERROR': pybasic.str_join('DISKDECOMP_UNCERTAINTY', pp_key_1, pp_key_2, separator='_', uppercase=True),
            'DISK_R_ERROR': pybasic.str_join('DISK', 'RADIANCE', 'UNCERTAINTY', pp_key_1, pp_key_2,
                                                  separator='_', uppercase=True),
            'DISK_R_RECT_ERROR': pybasic.str_join('DISK', 'RECTIFIED', 'RADIANCE', 'UNCERTAINTY', pp_key_1, pp_key_2,
                                                  separator='_', uppercase=True),
            'DISK_CALIB_ERROR': pybasic.str_join('DISK', 'CALIBRATION', 'UNCERTAINTY', pp_key_1, pp_key_2,
                                                  separator='_', uppercase=True),
            'DQI': pybasic.str_join('DQI', pp_key_1, pp_key_2, 'CHAN', separator='_', uppercase=True),
        }

        for vnp, vn_nc in dict_vn_nc.items():
            disk_array = np.array(dataset.variables[vn_nc])
            for ind, emline in enumerate(variables['EMISSION_SPECTRA']):
                if 'ERROR' in vnp:
                    vn = vnp.replace('_ERROR', '') +  '_' + emline + '_ERROR'
                else:
                    vn = vnp + '_' + emline

                variables[vn] = disk_array[:, :, ind].T[ind_along, :]

        dataset.close()

        self.variables = variables
        self.metadata = metadata



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import numpy as np
import datetime
import copy

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, InstrumentModel, ProductModel
from geospacelab.datahub.sources.esa_eo import esaeo_database
from geospacelab.datahub.sources.esa_eo.swarm import swarm_facility
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool
from geospacelab.datahub.sources.esa_eo.swarm.advanced.efi_lp_hm.loader import Loader as default_Loader
from geospacelab.datahub.sources.esa_eo.swarm.advanced.efi_lp_hm.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.esa_eo.swarm.advanced.efi_lp_hm.variable_config as var_config


default_dataset_attrs = {
    'database': esaeo_database,
    'facility': swarm_facility,
    'instrument': 'EFI-LP',
    'product': 'LP_HM',
    'data_file_ext': 'cdf',
    'product_version': 'latest',
    'data_root_dir': prf.datahub_data_root_dir / 'ESA' / 'SWARM' / 'Advanced',
    'allow_load': True,
    'allow_download': True,
    'force_download': False,
    'data_search_recursive': False,
    'add_AACGM': False,
    'add_APEX': False,
    'quality_control': False,
    'calib_control': False,
    'label_fields': ['database', 'facility', 'instrument', 'product'],
    'load_mode': 'AUTO',
    'time_clip': True,
}

default_variable_names = [
    'SC_DATETIME',
    'SC_GEO_LAT',
    'SC_GEO_LON',
    'SC_GEO_ALT',
    'SC_GEO_r',
    'SC_SZA',
    'SC_SAz',
    'SC_ST',
    'SC_DIP_LAT',
    'SC_DIP_LON',
    'SC_QD_MLT',
    'SC_QD_LAT',
    'SC_AACGM_LAT',
    'SC_AACGM_LON',
    'n_e',
    'T_e_HGN',
    'T_e_LGN',
    'T_e',
    'V_s_HGN',
    'V_s_LGN',
    'SC_U',
    'FLAG'
    ]

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', 'ESA/EarthOnline')
        self.facility = kwargs.pop('facility', 'SWARM')
        self.instrument = kwargs.pop('instrument', 'EFI-LP')
        self.product = kwargs.pop('product', 'HM02')
        self.product_version = kwargs.pop('product_version', '')
        self.local_latest_version = ''
        self.allow_download = kwargs.pop('allow_download', False)
        self.force_download = kwargs.pop('force_download', False)
        self.quality_control = kwargs.pop('quality_control', False)
        self.calib_control = kwargs.pop('calib_control', False)
        self.add_AACGM = kwargs.pop('add_AACGM', False)
        self.add_APEX = kwargs.pop('add_APEX', False)
        self._data_root_dir_init = copy.deepcopy(self.data_root_dir)   # Record the initial root dir

        self.sat_id = kwargs.pop('sat_id', 'A')

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not list(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        self.data_root_dir = self.data_root_dir / self.instrument / self.product

        if str(self.product_version) and self.product_version != 'latest':
            self.data_root_dir = self.data_root_dir / self.product_version
        else:
            self.product_version = 'latest'
            self.force_download = False
            mylog.simpleinfo.info(f'Checking the latest version of the data file ...')
            self.download_data()
            
            # try:
            #     dirs_product_version = [f.name for f in self.data_root_dir.iterdir() if f.is_dir()]
            # except FileNotFoundError:
            #     dirs_product_version = []
            #     self.force_download = True
            # else:
            #     if not list(dirs_product_version):
            #         self.force_download = True

            # if list(dirs_product_version):
            #     self.local_latest_version = max(dirs_product_version)
            #     self.data_root_dir = self.data_root_dir / self.local_latest_version
            #     if not self.force_download:
            #         mylog.simpleinfo.info(
            #             "Note: Loading the local files " +
            #             "with the latest version {} ".format(self.local_latest_version) +
            #             "Keep an eye on the latest baselines online!"
            #         )


    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )
        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type='cdf')

            for var_name in self._variables.keys():
                value = load_obj.variables[var_name]
                self._variables[var_name].join(value)

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range(var_datetime_name='SC_DATETIME')
        if self.quality_control:
            self.time_filter_by_quality()
        if self.calib_control:
            self.time_filter_by_calib()
            
        if self.add_AACGM:
            self.convert_to_AACGM()

        if self.add_APEX:
            self.convert_to_APEX()

        self.add_GEO_LST()

    def add_GEO_LST(self):
        lons = self['SC_GEO_LON'].flatten()
        uts = self['SC_DATETIME'].flatten()
        lsts = [ut + datetime.timedelta(hours=lon / 15.) for ut, lon in zip(uts, lons)]
        lsts = [lst.hour + lst.minute / 60. + lst.second / 3600. for lst in lsts]
        var = self.add_variable(var_name='SC_GEO_LST')
        var.value = np.array(lsts)[:, np.newaxis]
        var.label = 'LST'
        var.unit = 'h'
        var.depends = self['SC_GEO_LON'].depends
        return var
    
    def convert_to_APEX(self):
        import geospacelab.cs as gsl_cs

        coords_in = {
            'lat': self['SC_GEO_LAT'].value.flatten(),
            'lon': self['SC_GEO_LON'].value.flatten(),
            'r': self['SC_GEO_r'].value.flatten() / 6371.2
        }
        dts = self['SC_DATETIME'].value.flatten()
        cs_sph = gsl_cs.GEOCSpherical(coords=coords_in, ut=dts)
        cs_apex = cs_sph.to_APEX(append_mlt=True)
        self.add_variable('SC_APEX_LAT')
        self.add_variable('SC_APEX_LON')
        self.add_variable('SC_APEX_MLT')
        self['SC_APEX_LAT'].value = cs_apex['lat'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_APEX_LON'].value = cs_apex['lon'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_APEX_MLT'].value = cs_apex['mlt'].reshape(self['SC_DATETIME'].value.shape)

    def convert_to_AACGM(self):
        import geospacelab.cs as gsl_cs

        coords_in = {
            'lat': self['SC_GEO_LAT'].value.flatten(),
            'lon': self['SC_GEO_LON'].value.flatten(),
            'r': self['SC_GEO_r'].value.flatten() / 6371.2
        }
        dts = self['SC_DATETIME'].value.flatten()
        cs_sph = gsl_cs.GEOCSpherical(coords=coords_in, ut=dts)
        cs_aacgm = cs_sph.to_AACGM(append_mlt=True)
        self.add_variable('SC_AACGM_LAT')
        self.add_variable('SC_AACGM_LON')
        self.add_variable('SC_AACGM_MLT')
        self['SC_AACGM_LAT'].value = cs_aacgm['lat'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_AACGM_LON'].value = cs_aacgm['lon'].reshape(self['SC_DATETIME'].value.shape)
        self['SC_AACGM_MLT'].value = cs_aacgm['mlt'].reshape(self['SC_DATETIME'].value.shape)

    def time_filter_by_quality(self, quality_flags=None):
        if quality_flags is None:
            quality_flags = np.array([1])

        for qf in quality_flags:
            inds = np.where(self['QUALITY_FLAG'].value.flatten() == qf)[0]
            for key in self.keys():
                self._variables[key].value = self._variables[key].value[inds, ::]

    def time_filter_by_calib(self, calib_flags=None):

        if calib_flags is None:
            calib_flags = np.array([0])

        for cf in calib_flags:
            inds = np.where(self['CALIB_FLAG'].value.flatten() == cf)[0]
            for key in self.keys():
                self._variables[key].value = self._variables[key].value[inds, ::]

    def search_data_files(self, **kwargs):

        dt_fr = self.dt_fr
        dt_to = self.dt_to

        diff_days = dttool.get_diff_days(dt_fr, dt_to)

        dt0 = dttool.get_start_of_the_day(dt_fr)

        for i in range(diff_days + 1):
            this_day = dt0 + datetime.timedelta(days=i)

            initial_file_dir = kwargs.pop(
                'initial_file_dir', self.data_root_dir
            )

            initial_file_dir = initial_file_dir / 'Sat_{}'.format(self.sat_id) / this_day.strftime("%Y")

            file_patterns = [
                'EFI' + self.sat_id.upper(),
                self.product.upper(),
                this_day.strftime('%Y%m%d') + 'T',
            ]
            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]
            search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir,
                search_pattern=search_pattern,
                allow_multiple_files=True,
            )
            # Validate file paths

            if (not done and self.allow_download) or self.force_download:
                done = self.download_data()
                if done:
                    initial_file_dir = self.data_root_dir
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir,
                        search_pattern=search_pattern,
                        allow_multiple_files=True
                    )

        return done

    def download_data(self, dt_fr=None, dt_to=None):
        if dt_fr is None:
            dt_fr = self.dt_fr
        if dt_to is None:
            dt_to = self.dt_to
        download_obj = self.downloader(
            dt_fr, dt_to,
            sat_id=self.sat_id,
            data_type=self.product,
            file_version=self.product_version,
            force=self.force_download
        )
        if download_obj.done:
            self.force_download = False

            if download_obj.file_version != self.local_latest_version and self.product_version == 'latest':
                mylog.simpleinfo.warning(
                    f"NOTE: The data with the latest version ({download_obj.file_version}) have been downloaded"
                )
            self.product_version = download_obj.file_version
            self.data_root_dir = copy.deepcopy(self._data_root_dir_init)
            self._validate_attrs()

        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        if isinstance(value, str):
            self._instrument = InstrumentModel(value)
        elif issubclass(value.__class__, InstrumentModel):
            self._instrument = value
        else:
            raise TypeError



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import requests
import bs4
import re
import pathlib

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf

from geospacelab.datahub.sources.cdaweb.downloader import CDAWebHTTPDownloader as DownloaderBase


class Downloader(DownloaderBase):
    
    def __init__(
        self, 
        dt_fr,  dt_to, 
        time_res='1min', 
        product='OMNI2', 
        version='',
        root_dir_local=None,
        direct_download=True,
        force_download=False,
        dry_run=False,
        ):

        self.time_res = time_res
        self.product = product
        self.version = version
        
        if root_dir_local is None:
            root_dir_local = prf.datahub_data_root_dir / "CDAWeb" / 'OMNI' 
        else:
            root_dir_local = pathlib.Path(root_dir_local)
        root_dir_remote = '/'.join([
            self.root_dir_remote, 'omni', 'omni_cdaweb'
            ])
        super().__init__(
            dt_fr, dt_to,
            root_dir_local=root_dir_local,
            root_dir_remote=root_dir_remote,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
        ) 
        
    def search_from_http(self, *args, **kwargs):
        dt_fr_1 = self.dt_fr
        dt_to_1 = self.dt_to
        diff_months = (dt_to_1.year - dt_fr_1.year) * 12 + dt_to_1.month - dt_fr_1.month
        file_paths_remote = []
        for nm in range(diff_months + 1):
            this_month = dttool.get_next_n_months(dt_fr_1, nm)
            
            subdirs = []
            
            if self.time_res in ['1min', '5min']:
                subdirs.append(f'{omni_product_dict[self.product]}_{self.time_res}')
            elif self.time_res == '1h':
                subdirs.append('hourly')
            subdirs.append('{:4d}'.format(this_month.year))
            
            if self.time_res in ['1min', '5min',]:
                file_name_patterns = [
                    'omni', omni_product_dict[self.product], self.time_res, this_month.strftime("%Y%m")
                    ]
            elif self.time_res == '1h':
                file_name_patterns = [
                    'omni2', 'mrg1hr', this_month.strftime("%Y")
                    ]
            if str(self.version):
                file_name_patterns.append(self.version)
                
            paths = super().search_from_http(subdirs=subdirs, file_name_patterns=file_name_patterns)
            
            if len(paths) > 1 and self.time_res != '1h':
                mylog.StreamLogger.error("Find multiple matched files!")
                print(paths)

            file_paths_remote.extend(paths)
        
        return list(set(file_paths_remote))
    
    def save_files_from_http(self, file_paths_local=None, root_dir_remote=None):
        
        if file_paths_local is None:
            file_paths_local = []
            for fp_remote in self.file_paths_remote:
                if self.time_res in ['1min', '5min']:
                    subdir = f'{self.product}_high_res_{self.time_res}'
                elif self.time_res == '1h':
                    subdir = 'OMNI2_low_res_1h'
            
                fp_local = fp_remote.replace(self.base_url + '/' + self.root_dir_remote + '/', '')
                pattern_replaced = fp_local.split('/')[0]
                fp_local = fp_local.replace(
                    pattern_replaced, subdir, 1
                )
                file_paths_local.append(self.root_dir_local / fp_local)
        return super().save_files_from_http(file_paths_local, root_dir_remote)
        

omni_product_dict = {
    'OMNI': 'hro',
    'OMNI2': 'hro2',
}


def test():
    dt_fr = datetime.datetime(2020, 3, 4)
    dt_to = datetime.datetime(2020, 5, 3)
    download_obj = Downloader(dt_fr, dt_to, force_download=True, time_res='1h')
    pass


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLAB (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import netCDF4
import datetime
import numpy as np
import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader(object):

    def __init__(self, file_path, file_type='grd', pole='N'):

        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type
        self.pole = pole

        self.load_data()

    def load_data(self):
        dataset = netCDF4.Dataset(self.file_path)
        variables = {}

        year = dataset.variables['year'][::][0]
        start_of_year = datetime.datetime(year, 1, 1) 
        hhs = [int(np.floor(t)) for t in dataset.variables['time'][::]]
        mms = [int(np.round((t-hh)*60)) for t, hh in zip(dataset.variables['time'][::], hhs)] 
        sss = [int(np.round((t-hh - mm/60)*60)) for t, hh, mm in zip(dataset.variables['time'][::], hhs, mms)]  
        dts = np.array([
            start_of_year + datetime.timedelta(days=int(doy)-1, hours=hh, minutes=mm, seconds=ss)
            for doy, hh, mm, ss in zip(
                dataset.variables['doy'][::],
                hhs,
                mms,
                sss,
            )
        ])

        ntime = dts.shape[0]
        nlon = dataset.variables['nLonGrid'][::][0]
        nlat = dataset.variables['nLatGrid'][::][0]
        colat = dataset.variables['cLat_deg'][::].reshape((ntime, nlon, nlat))

        mlt = dataset.variables['mlt_hr'][::].reshape((ntime, nlon, nlat))

        Jr = dataset.variables['jPar'][::].reshape(ntime, nlon, nlat)

        variables['DATETIME'] = dts[:, np.newaxis]

        variables['GRID_MLAT'] = np.array(90. - colat)
        variables['GRID_MLT'] = np.array(mlt)

        variables['GRID_Jr'] = np.array(Jr)

        dataset.close()

        self.variables = variables


cdf_var_names = [
    'npnt', 'year', 'doy', 'time', 'avgint',
    'kmax', 'mmax', 'res_deg', 'nLatGrid', 'nLonGrid',
    'cLat_deg', 'mlt_hr', 'geo_cLat_deg', 'geo_lon_deg', 'R',
    'pos_geo', 'db_R', 'db_T', 'db_P', 'db_geo',
    'jPar', 'db_Th_Th', 'db_Ph_Th', 'db_Th_Ph', 'db_Ph_Ph',
    'del_db_R', 'del_db_T', 'del_db_P', 'del_db_geo', 'del_jPar',
    'del_db_Th_Th', 'del_db_Ph_Th', 'del_db_Th_Ph', 'del_db_Ph_Ph'
]




if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path('/home/lei/git-repos/geospacelab/geospacelab/datahub/sources/jhuapl/ampere/20160314/AMPERE_GRD_20160314T0000_20160314T0100_N.nc')
    # fp = "/home/lei/north.nc"
    loader = Loader(file_path=fp)




    # if hasattr(readObj, 'pole'):
    #    readObj.filter_data_pole(boundinglat = 25)

# dataset.variables
# {'npnt': <class 'netCDF4._netCDF4.Variable'>
# int16 npnt(nRec)
#     comment: number of observations in record.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'year': <class 'netCDF4._netCDF4.Variable'>
# int16 year(nRec)
#     comment: year.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'doy': <class 'netCDF4._netCDF4.Variable'>
# int16 doy(nRec)
#     comment: day of year (1-366).
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'time': <class 'netCDF4._netCDF4.Variable'>
# float32 time(nRec)
#     comment: time in units of fractional hours of day.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'avgint': <class 'netCDF4._netCDF4.Variable'>
# int16 avgint(nRec)
#     comment: time averaging window length in seconds.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'kmax': <class 'netCDF4._netCDF4.Variable'>
# int16 kmax(nRec)
#     comment: latitude order of fit.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'mmax': <class 'netCDF4._netCDF4.Variable'>
# int16 mmax(nRec)
#     comment: longitude order of fit.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'res_deg': <class 'netCDF4._netCDF4.Variable'>
# float32 res_deg(nRec)
#     comment: grid latitude resolution in degrees.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'nLatGrid': <class 'netCDF4._netCDF4.Variable'>
# int16 nLatGrid(nRec)
#     comment: number of latitude points in grid.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'nLonGrid': <class 'netCDF4._netCDF4.Variable'>
# int16 nLonGrid(nRec)
#     comment: number of longitude points in grid.
# unlimited dimensions:
# current shape = (720,)
# filling on, default _FillValue of -32767 used, 'cLat_deg': <class 'netCDF4._netCDF4.Variable'>
# float32 cLat_deg(nRec, nObs)
#     comment: co-latitude in AACGM coordinates in degrees.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'mlt_hr': <class 'netCDF4._netCDF4.Variable'>
# float32 mlt_hr(nRec, nObs)
#     comment: AACGM Magnetic Local Time (MLT) in hours.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'geo_cLat_deg': <class 'netCDF4._netCDF4.Variable'>
# float32 geo_cLat_deg(nRec, nObs)
#     comment: co-latitude in GEO coordinates in degrees.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'geo_lon_deg': <class 'netCDF4._netCDF4.Variable'>
# float32 geo_lon_deg(nRec, nObs)
#     comment: longitude in GEO coordinates in degrees.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'R': <class 'netCDF4._netCDF4.Variable'>
# float32 R(nRec, nObs)
#     comment: Radius from center of the Earth in kilometers.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'pos_geo': <class 'netCDF4._netCDF4.Variable'>
# float32 pos_geo(nRec, nObs, vComp)
#     comment: vehicle position in GEO coordinates in units of kilometers.
# unlimited dimensions:
# current shape = (720, 1200, 3)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_R': <class 'netCDF4._netCDF4.Variable'>
# float64 db_R(nRec, nObs)
#     comment: magnetic field perturbation parallel to GEO radial direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_T': <class 'netCDF4._netCDF4.Variable'>
# float64 db_T(nRec, nObs)
#     comment: magnetic field perturbation parallel to GEO northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_P': <class 'netCDF4._netCDF4.Variable'>
# float64 db_P(nRec, nObs)
#     comment: magnetic field perturbation parallel to GEO eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_geo': <class 'netCDF4._netCDF4.Variable'>
# float64 db_geo(nRec, nObs, vComp)
# unlimited dimensions:
# current shape = (720, 1200, 3)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'jPar': <class 'netCDF4._netCDF4.Variable'>
# float64 jPar(nRec, nObs)
#     comment: Radial current density [muA/m^2]
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_Th_Th': <class 'netCDF4._netCDF4.Variable'>
# float64 db_Th_Th(nRec, nObs)
#     comment: Magnetic field perturbation parallel to the AACGM northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_Ph_Th': <class 'netCDF4._netCDF4.Variable'>
# float64 db_Ph_Th(nRec, nObs)
#     comment: Magnetic field perturbation perpendicular to the AACGM northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_Th_Ph': <class 'netCDF4._netCDF4.Variable'>
# float64 db_Th_Ph(nRec, nObs)
#     comment: Magnetic field perturbation perpendicular  to the AACGM eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'db_Ph_Ph': <class 'netCDF4._netCDF4.Variable'>
# float64 db_Ph_Ph(nRec, nObs)
#     comment: Magnetic field perturbation parallel to the AACGM eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_R': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_R(nRec, nObs)
#     comment: magnetic field residual parallel to GEO radial direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_T': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_T(nRec, nObs)
#     comment: magnetic field residual parallel to GEO northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_P': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_P(nRec, nObs)
#     comment: magnetic field residual parallel to GEO eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_geo': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_geo(nRec, nObs, vComp)
#     comment: magnetic field residuals in GEO coordinates in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200, 3)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_jPar': <class 'netCDF4._netCDF4.Variable'>
# float64 del_jPar(nRec, nObs)
#     comment: Radial current density residual [muA/m^2]
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_Th_Th': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_Th_Th(nRec, nObs)
#     comment: Magnetic field residual parallel to the AACGM northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_Ph_Th': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_Ph_Th(nRec, nObs)
#     comment: Magnetic field residual perpendicular to the AACGM northward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_Th_Ph': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_Th_Ph(nRec, nObs)
#     comment: Magnetic field residual perpendicular  to the AACGM eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used, 'del_db_Ph_Ph': <class 'netCDF4._netCDF4.Variable'>
# float64 del_db_Ph_Ph(nRec, nObs)
#     comment: Magnetic field residual parallel to the AACGM eastward direction in units of nano-Tesla.
# unlimited dimensions:
# current shape = (720, 1200)
# filling on, default _FillValue of 9.969209968386869e+36 used}



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, ProductModel

from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog

from geospacelab.datahub.sources.gfz import gfz_database
from geospacelab.datahub.sources.gfz.kpap.loader import Loader as default_Loader
from geospacelab.datahub.sources.gfz.kpap.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.gfz.kpap.variable_config as var_config


default_dataset_attrs = {
    'database': gfz_database,
    'product': 'KpAp',
    'data_file_ext': 'nc',
    'data_root_dir': prf.datahub_data_root_dir / 'GFZ' / 'Indices',
    'allow_load': True,
    'allow_download': True,
    'force_download': False,
    'data_search_recursive': False,
    'label_fields': ['database', 'product', 'data_file_ext'],
    'time_clip': True,
}

default_variable_names = ['DATETIME', 'Kp', 'ap', 'Ap']

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', gfz_database)
        self.product = kwargs.pop('product', 'KpAp')
        self.allow_download = kwargs.pop('allow_download', True)
        self.force_download = kwargs.pop('force_download', True)

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type=self.data_file_ext)

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def search_data_files(self, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_years = dt_to.year - dt_fr.year
        dt0 = datetime.datetime(dt_fr.year, 1, 1)
        for i in range(diff_years + 1):
            thisyear = datetime.datetime(dt0.year + i, 1, 1)
            if datetime.date.today().year == thisyear.year:
                self.force_download = True

            initial_file_dir = kwargs.pop('initial_file_dir', None)
            if initial_file_dir is None:
                initial_file_dir = self.data_root_dir / 'Kp_Ap'
            file_patterns = [thisyear.strftime("%Y")]
            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            if not self.force_download:
                done = super().search_data_files(
                    initial_file_dir=initial_file_dir, search_pattern=search_pattern
                )
            else:
                done = False

            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern)
                else:
                    print('Cannot find files from the online database!')

        return done

    def download_data(self):
        if self.data_file_ext == 'nc':
            download_obj = self.downloader(self.dt_fr, self.dt_to, data_file_root_dir=self.data_root_dir, force=self.force_download)
        else:
            raise NotImplementedError
        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

















# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy as np


def isnumeric(v):
    v_type = type(v)

    if np.issubdtype(v_type, np.integer) or np.issubdtype(v_type, np.floating):
        return True
    else:
        return False


def input_with_default(prompt, default=''):
    ip = input(prompt) or default
    return ip


def dict_set_default(dict1, *args, **kwargs):
    for ind, arg in enumerate(args):
        if ind % 2 == 0:
            kwargs[arg] = args[ind+1]
        else:
            continue
    for key, value in kwargs.items():
        dict1.setdefault(key, value)
    return dict1


def dict_print_tree(dict_in, level=0, max_level=None, value_repr=True, dict_repr=True, full_value=False):
    if level == 0:
        print("\x1b[0;31;40m" + retrieve_name(dict_in) + "\x1b[0m", end="")

    try:
        keys = dict_in.keys()
    except AttributeError:

        if value_repr:
            print("\x1b[2;37;40m" + ": " + "\x1b[0m", end="")
            print("\x1b[0;30;47m" + repr(dict_in) + "\x1b[0m")
        else:
            print("")

        if full_value:
            try:
                print(dict_in[:])
            except:
                print("Not implemented!")
        return None

    if dict_repr:
        print("\x1b[2;37;40m" + ": " + "\x1b[0m", end="")
        print("\x1b[7;34;40m" + repr(dict_in) + "\x1b[0m", end="")
    print("")
    if max_level is not None:
        if level >= max_level:
            return None

    for key in keys:
        print("\x1b[2;37;40m" + "|---" * (level + 1) + "\x1b[0m", end="")
        print("\x1b[1;33;40m" + key + "\x1b[0m", end="")
        new_dict = dict_in[key]
        dict_print_tree(new_dict, level+1, max_level=max_level,
                        value_repr=value_repr, dict_repr=dict_repr, full_value=full_value)


def retrieve_name(var):
    """
    Gets the name of var. Does it from the out most frame inner-wards.
    :param var: variable to get name from.
    :return: string
    """
    import inspect
    for fi in reversed(inspect.stack()):
        names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is var]
        if len(names) > 0:
            return names[0]


def str_join(*args, separator='_', uppercase=False, lowercase=False):
    """
    Join multiple strings into one. The empty string '' will be ignored.
    """
    if uppercase and lowercase:
        raise KeyError("The keywords 'uppercase' and 'lowercase' cannot be set as True at the same time!")

    strList_new = []
    for elem in args:
        if elem == '' or elem is None:
            continue
        if uppercase:
            elem = elem.upper()
        if lowercase:
            elem = elem.lower()
        strList_new.append(elem)
    return separator.join(strList_new)


def list_flatten(l_in, l_out=None):
    if l_out is None:
        l_out = []
    for ind, elem in enumerate(l_in):
        if type(elem) == list:
            list_flatten(l_in, l_out=l_out)
        else:
            l_out.append(elem)
    return l_out



import datetime
import pathlib
import copy

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf
from geospacelab.datahub.sources.cdaweb.downloader import CDAWebHTTPDownloader as DownloaderBase


class Downloader(DownloaderBase):
    
    def __init__(
            self,
            dt_fr=None, dt_to=None,
            sat_id=None,
            orbit_id=None,
            product='EDR_AUR',
            direct_download=True,
            force_download=False,
            dry_run=False,
            root_dir_local = None,
        ):
         
        self.sat_id = sat_id
        self.orbit_id = orbit_id
        self.product = product
        if root_dir_local is None:
            root_dir_local = prf.datahub_data_root_dir / 'CDAWeb' / 'DMSP' / 'SSUSI' / self.product.upper
        else:
            root_dir_local = pathlib.Path(root_dir_local)
        root_dir_remote = '/'.join([
            self.root_dir_remote, 'dmsp', 'dmsp'+self.sat_id.lower(), 'ssusi', 'data', product_dict[self.product]
            ])
        super().__init__(
            dt_fr, dt_to,
            root_dir_local=root_dir_local,
            root_dir_remote=root_dir_remote,
            direct_download=direct_download,
            force_download=force_download,
            dry_run=dry_run,
        ) 


    def search_from_http(self, *args, **kwargs):

        dt_fr_1 = self.dt_fr - datetime.timedelta(hours=3)
        dt_to_1 = self.dt_to + datetime.timedelta(hours=3)
        diff_days = dttool.get_diff_days(dt_fr_1, dt_to_1)
        dt0 = dttool.get_start_of_the_day(dt_fr_1)
        file_paths_remote = []
        for nd in range(diff_days + 1):
            this_day = dt0 + datetime.timedelta(days=nd)
            doy = dttool.get_doy(this_day)
            sdoy = '{:03d}'.format(doy)
            subdirs = [str(this_day.year), sdoy]

            if self.orbit_id is None:
                file_name_patterns = [
                    'dmsp' + self.sat_id.lower(),
                    'ssusi',
                    product_dict[self.product],
                    this_day.strftime("%Y") + sdoy + 'T',
                    '.nc'
                ]
            else:
                file_name_patterns = [
                    'dmsp' + self.sat_id.lower(),
                    'ssusi',
                    product_dict[self.product],
                    'REV',
                    self.orbit_id,
                    '.nc'
                ]
            paths = super().search_from_http(subdirs=subdirs, file_name_patterns=file_name_patterns)
            file_paths_remote.extend(paths)
        return file_paths_remote
    
    def save_files_from_http(self, file_paths_local=None, root_dir_remote=None):
        if file_paths_local is None:
            file_paths_local = []
            for fp_remote in self.file_paths_remote:
                sy = fp_remote.split('/')[-3]
                sdoy = fp_remote.split('/')[-2]
                year = int(sy)
                this_day = dttool.convert_doy_to_datetime(year, int(sdoy))
                file_dir_local = self.root_dir_local /  self.sat_id.upper() / sy / this_day.strftime("%Y%m%d")
                file_paths_local.append(file_dir_local / fp_remote.split('/')[-1])
        return super().save_files_from_http(file_paths_local, root_dir_remote)
                

product_dict = {
    'EDR_AUR': 'edr-aurora',
    'EDR_DAY_LIMB': 'edr-day-limb',
    'EDR_IONO': 'edr-iono',
    'EDR_NIGHT_LIMB': 'edr-night-limb',
    'L1B': 'l1b',
    'SDR_DISK': 'sdr-disk',
    'SDR_LIMB': 'sdr-limb',
    'SDR2_DISK': 'sdr2-disk',
    'SPECT_L1B': 'spect-l1b',
}

